package org.wakeland.pedro;

import java.io.IOException;
import java.io.OutputStream;
import java.util.Calendar;
import java.util.Date;

import javax.microedition.io.Connector;
import javax.microedition.io.file.FileConnection;
import javax.microedition.lcdui.*;
import javax.microedition.media.*;
import javax.microedition.media.control.VideoControl;

/**
 * 
 * J2ME class for capturing images. It enables to show last captured image in the form, 
 * save last captured image to the Images directory and send the last captured image
 * to the server (J2SE application).
 * 
 * <p>Original sample code of camera MIDlet downloaded from
 * <a href="http://wiki.forum.nokia.com/index.php/How_to_display_the_camera_%26_take_snapshots_in_Java_ME">Nokia wiki forum</a>.</p>
 * 
 * @author Peter (Pedro) Nosal, pedron4@gmail.com
 * @version 1.0
 * 
 */
public class Camera implements CommandListener {

	/**
	 * Instance of main class - MIDlet.
	 */
	private MobileImagineConnectorClient midlet;
	/**
	 * Form where captured image is placed.
	 */
    private Form formImages;
	/**
	 * Form where camera viewfinder is placed.
	 */
    private Form formCamera;
    /**
     * Command for going back to MIDlet. 
     */
    private Command cmdBackToMidlet;
    /**
     * Command for changing current screen to the form with last captured image. 
     */
    private Command cmdBackToImages;
    /**
     * Command for capturing image. 
     */
    private Command cmdCapture;
    /**
     * Command for changing current screen to the form with camera viewfinder. 
     */
    private Command cmdCamera;
    /**
     * Command for sending captured image to server.
     */
    private Command cmdSendImageToServer;
    /**
     * Command for saving captured image. 
     */
    private Command cmdSave;
    /**
     * Player for camera.
     */
    private Player player;
    /**
     * Video control of camera.
     */
    private VideoControl videoControl;
    /**
     * Class extending thread which runs capturing of image.
     */
    private Video video;
    /**
	 * Alert to be displayed if error occurs or to show info message.
     */
    private Alert alert;
    /**
     * Array of bytes for storing last captured image.
     */
    byte[] imageRaw = null;

    /**
     * Constructs new Camera class.
     * Creates all commands and forms used in this class and checks camera support.
     * @param midlet	Main class - MIDlet.
     */
    public Camera(MobileImagineConnectorClient midlet) {
    	this.midlet = midlet;

    	if (checkCameraSupport() == false) {
            showAlert("Alert", "Camera is not supported!", AlertType.ERROR, midlet.getFormMain());
            return;
        }

    	cmdCamera = new Command("Camera", Command.SCREEN, 0);
    	cmdBackToImages = new Command("Images", Command.BACK, 0);
    	cmdBackToMidlet = new Command("Back", Command.BACK, 0);
    	cmdCapture = new Command("Capture", Command.SCREEN, 0);
    	cmdSendImageToServer = new Command("Send", Command.SCREEN, 0);
    	cmdSave = new Command("Save", Command.SCREEN, 0);
 
        formImages = new Form("Captured image");
        formImages.addCommand(cmdCamera);
        formImages.addCommand(cmdBackToMidlet);
        formImages.setCommandListener(this);

        formCamera = new Form("Camera");
        formCamera.addCommand(cmdCapture);
        formCamera.addCommand(cmdBackToImages);
        formCamera.setCommandListener(this);


        formImages.append("There are no captured images!\n");
    }

    /**
     * Returns camera form.
     * @return	Camera form.
     */
	Form getFormCamera(){
    	return formCamera;
    }

    /**
     * Returns images form.
     * @return	Images form.
     */
    Form getFormImages(){
    	return formImages;
    }

    /**
     * Captures image.
     */
    void capture(){
    	imageRaw = null;
        video = new Video();
        video.start();
    }

    /**
     * Returns whether some image has been captured yet.
     * @return	True if some image has been captured; false otherwise.
     */
    boolean isCaptured(){
    	return imageRaw != null;
    }

    /**
     * Switches a current displayable in a display. Calls setMIDletDisplay()
     * method in ImageMobileConnectorClient with the same parameters.
     * @param alert 			The Alert which is temporarily set to the display.
     * @param nextDisplayable 	The Displayable to be set.
     */
    private void setDisplay(Alert alert, Displayable nextDisplayable) {
    	midlet.setMIDletDisplay(alert, nextDisplayable);
    }

    /**
     * Shows alert with specified title, text and type.
     * @param title 			Title of alert.
     * @param message 			Text of alert.
     * @param type				Type of alert.
     * @param nextDisplayable 	Next displayable.
     */
    private void showAlert(String title, String message, AlertType type, Displayable nextDisplayable) {
        alert = new Alert(title, message, null, type);
        alert.setTimeout(Alert.FOREVER);
        alert.setCommandListener(this);

        midlet.setMIDletDisplay(alert, nextDisplayable);
    }        

    /**
     * Checks camera support.
     * @return 		True if camera is supported, false otherwise.
     */
    private boolean checkCameraSupport() {
        String propValue = System.getProperty("supports.video.capture");
        return (propValue != null) && propValue.equals("true");
    }    

    /**
     * From CommandListener:
     * Called by a system indicating that a command has been invoked.
     * @param command 		The Command that was invoked.
     * @param displayable 	The Displayable where the command was invoked.
     */
    public void commandAction(Command command, Displayable displayable) {
    	if (command == cmdCamera) {
            showCamera();
        } else if (command == cmdBackToImages) {
        	if (player != null){
	            player.close();
	            player = null;
        	}
            videoControl = null;

        	setDisplay(null, formImages);
        } else if (command == cmdBackToMidlet) {
        	if (player != null){
	            player.close();
	            player = null;
        	}
            videoControl = null;

            setDisplay(null, midlet.getFormMain());
        } else if (command == cmdCapture) {
        	capture();
        } else if (command == cmdSendImageToServer){
        	sendImageToServer();
        } else if (command == cmdSave){
        	saveImage();
        }
    }

    /**
     * Sends the last captured image to server.
     */
    void sendImageToServer(){
		midlet.getObexClient();
		midlet.sendMessageToServer(imageRaw, ObexClient.MESSAGE_TYPE_IMAGE);
	}

    /**
     * Returns current minute, hour, day, month and year in format YYYY-MM-DD_HH-MM.
     * @return		Current minute, hour, day, month and year in format YYYY-MM-DD_HH-MM.
     */
    private String getDateTime(){
        Calendar calendar = Calendar.getInstance();
        calendar.setTime(new Date());
        String year = String.valueOf(calendar.get(Calendar.YEAR));
        String month = String.valueOf(calendar.get(Calendar.MONTH)+1);
        if (month.length() == 1)
        	month = "0" + month;
        String date = String.valueOf(calendar.get(Calendar.DATE));
        if (date.length() == 1)
        	date = "0" + date;
        String hour = String.valueOf(calendar.get(Calendar.HOUR_OF_DAY ));
        if (hour.length() == 1)
        	hour = "0" + hour;
        String minute = String.valueOf(calendar.get(Calendar.MINUTE));
        if (minute.length() == 1)
        	minute = "0" + minute;
        return year + "-" + month + "-" + date + "_" + hour + "-" + minute;
    }

    /**
     * Saves the last captured image to Images directory.
     */
    private void saveImage(){
        FileConnection file = null;
        OutputStream outStream = null;

        // Get path to photos folder.
        String dirPhotos = System.getProperty("fileconn.dir.photos");
        if (dirPhotos == null) {
            showAlert("Error", "Unable get photos folder name", AlertType.ERROR, midlet.getFormMain());
        }

        String fileName = dirPhotos + "ImagineMobile_" + getDateTime() + ".jpeg";
        // Open file
        try {
			file = (FileConnection)Connector.open(fileName, Connector.READ_WRITE);
		} catch (IOException e) {
			showAlert("Error", "IOException on opening connector", AlertType.ERROR, midlet.getFormMain());
		}
        // If there is no file then create it.
        if (file.exists() == false) {
            try {
				file.create();
			} catch (IOException e) {
				showAlert("Error", "IOException on creating file", AlertType.ERROR, midlet.getFormMain());
			}
        }
        // Write data received from camera while making snapshot to file.
        try {
			outStream = file.openOutputStream();
		} catch (IOException e) {
			showAlert("Error", "IOException on opening output stream", AlertType.ERROR, midlet.getFormMain());
		}
        try {
			outStream.write(imageRaw);
		} catch (IOException e) {
			showAlert("Error", "IOException on writing to output stream", AlertType.ERROR, midlet.getFormMain());
		}

		// Try to close file.
        try {
            if (outStream != null) {
                outStream.close();
            }
            if (file != null) {
                file.close();
            }
        } catch (Exception exc) {
			showAlert("Error", "IOException on closing output stream", AlertType.ERROR, null);
        }

        showAlert("Info", "Image is saved in " + fileName, AlertType.INFO, formImages);
    }

    /**
     * Shows the form with the camera viewfinder.
     */
    void showCamera() {
        try {
        	// We make sure, that variable player is null.
        	// It should be null, but if it is not, MediaException would be thrown.
        	if (player != null){
	            player.close();
	            player = null;
        	}
            videoControl = null;

            player = Manager.createPlayer("capture://video"); // "capture://video" is used for S60 devices
            //player = Manager.createPlayer("capture://image"); // "capture://image" is used for Series 40 devices
            player.realize();
 
            videoControl = (VideoControl)player.getControl("VideoControl");
            formCamera.deleteAll();
            formCamera.append((Item)videoControl.initDisplayMode(VideoControl.USE_GUI_PRIMITIVE, null));
            setDisplay(null, formCamera);

            player.start();
        } catch (IOException ioe) {} catch (MediaException me) {
        	formImages.append("MediaException: " + me.toString() + "\n");
        }
    }

    /**
     * Class for capturing image in a camera viewfinder.
     * 
	 * <p>Original sample code of camera MIDlet downloaded from
	 * <a href="http://wiki.forum.nokia.com/index.php/How_to_display_the_camera_%26_take_snapshots_in_Java_ME">Nokia wiki forum</a>.</p>
	 * 
	 * @author Peter (Pedro) Nosal, pedron4@gmail.com
	 * @version 1.0
     *
     */
    class Video extends Thread {

    	/**
    	 * Calls the captureVideo method.
    	 */
    	public void run() {
            captureVideo();
        }

    	/**
    	 * Captures the image, shows it in the images form and 
    	 * switches displayable to images form.
    	 */
        private void captureVideo() {
            try {
                if (checkJpegEncodingSupport() == false) {
                	midlet.printToForm("jpeg is not supported");
                    showAlert("Error", "Jpeg encoding is not supported!", AlertType.ERROR, midlet.getFormMain());
                }

                // If there wasn't captured image yet, add "Send" and 
                // "Save" commands.
                if (!isCaptured()){
            		formImages.addCommand(cmdSendImageToServer);
            		formImages.addCommand(cmdSave);
            	}

            	imageRaw = videoControl.getSnapshot("encoding=jpeg"); //&colors=gray
                Image image = createScaledImage(Image.createImage(imageRaw, 0, imageRaw.length));
                ImageItem imageItem = new ImageItem(null, image, ImageItem.LAYOUT_DEFAULT, "Image can not be displayed.");
            	// deleting previous image
            	formImages.deleteAll();
            	// appending new image
                formImages.append(imageItem);

                setDisplay(null, formImages);
 
                player.close();
                player = null;
                videoControl = null;
            } catch (MediaException me) { }
        }

        /**
         * Method used to scale each image down to the size of the screen.
         * Adapted from <a href="http://www.liviutudor.com/tech/programming/mobile/j2me/#scale_image_to_screen">Liviutudor.com</a>
         *
         * @param image 	Image to scale
         * @return Scaled 	Version of the image, having the width and height equal to the size of the screen.
         */
        private Image createScaledImage(Image image) {
        	final int sourceWidth = image.getWidth();
        	final int sourceHeight = image.getHeight();
        	final int screenWidth = midlet.getDisplay().getCurrent().getWidth();
        	final int screenHeight = (int) (screenWidth * 0.8); //midlet.getDisplay().getCurrent().getHeight();
        	final Image sImg = Image.createImage( screenWidth, screenHeight );
        	final Graphics g = sImg.getGraphics();
        	final int[] lineRGB = new int[sourceWidth];
        	final int[] sourcePos = new int[screenWidth]; // cache for x positions
        	{
        		/*
        		 * Pre-calculate x positions with modified bresenham algorithm
        		 * http://www.cs.helsinki.fi/group/goa/mallinnus/lines/bresenh.html
        		 */
        		int y = 0;
        		int eps = -(sourceWidth >> 1);
        		for (int x = 0; x < sourceWidth; x++) {
        			eps += screenWidth;
        			if ((eps << 1) >= sourceWidth){
        				if (++y == screenWidth)
        					break;
        				sourcePos[y] = x;
        				eps -= sourceWidth;
        			}
        		}
        	}
        	for (int y = 0; y < screenHeight; y++ ) {
        		image.getRGB( lineRGB, 0, sourceWidth, 0, y * sourceHeight / screenHeight, sourceWidth, 1 );
        		for (int x = 1; x < screenWidth; x++ ) // skip pixel 0
        			lineRGB[x] = lineRGB[sourcePos[x]];
        		g.drawRGB( lineRGB, 0, screenWidth, 0, y, screenWidth, 1, false );
        	}
        	return sImg;
        }

        /**
         * Checks jpeg encoding support.
         * @return 		True if jpeg encoding is supported, false otherwise.
         */
        private boolean checkJpegEncodingSupport() {
            String encodings = System.getProperty("video.snapshot.encodings");
            return (encodings != null) && (encodings.indexOf("jpeg") != -1);
        }
    }
}