package org.wakeland.pedro;

import java.io.IOException;

import javax.bluetooth.RemoteDevice;
import javax.microedition.lcdui.Alert;
import javax.microedition.lcdui.AlertType;
import javax.microedition.lcdui.Canvas;
import javax.microedition.lcdui.Command;
import javax.microedition.lcdui.CommandListener;
import javax.microedition.lcdui.Display;
import javax.microedition.lcdui.Displayable;
import javax.microedition.lcdui.Form;
import javax.microedition.lcdui.List;
import javax.microedition.lcdui.TextBox;
import javax.microedition.lcdui.TextField;
import javax.microedition.midlet.MIDlet;
import javax.microedition.midlet.MIDletStateChangeException;

/**
 * 
 * Main class (MIDlet), for connecting to MobileImagineConnectorServer (MICS) 
 * (running on PC) over Bluetooth protocol OBEX. MICS is able to connect 
 * to Imagine Logo, so this MIDlet can communicate with Imagine Logo. This MIDlet 
 * is able to send/receive messages to/from Imagine (as chat), capture pictures
 * and save them to the phone's memory or send it to Imagine. This MIDlet also
 * uses a phone's keypad (game keys up, down, left and right) to control an NXT
 * brick through Imagine.
 * 
 * <p>This application together with MobileImagineConnectorServer is a bachelor's degree
 * project of Peter Nosal. There is a dedicated website to this bachelor theses:
 * <a href="http://pedro.wakeland.org/index.php?page=bc">pedro.wakeland.org</a> (only is Slovak).</p>
 * 
 * @author Peter (Pedro) Nosal, pedron4@gmail.com
 * @version 1.0
 * 
 */
public class MobileImagineConnectorClient extends MIDlet implements CommandListener {

	/**
	 * Instance of Display - for controlling of the output on the screen.
	 */
	private Display display;
    /**
     * Instance of Camera - for capturing pictures.
     */
	private Camera camera;
    /**
     * Instance of OBEXclient - for connecting to server.
     */
	private ObexClient obexClient;
    /**
     * Instance of Joystick - for controlling NXT robot.
     */
	private Canvas joystick;
	/**
	 * Main form object for MIDlet.
	 */
	private Form formMain;
	/**
	 * Form for chat.
	 */
	private Form formChat;
	/**
	 * Alert to be displayed if error occurs or to show info message.
	 */
	private Alert alert;
	/**
	 * Command for exiting from MIDlet.
	 */
	private Command cmdExit;
	/**
	 * Command for searching for Bluetooth devices.
	 */
	private Command cmdOBEXclient;
	/**
	 * Command for going back, mostly to main form.
	 */
	private Command cmdBack;
	/**
	 * Command for changing current screen to chat.
	 */
	private Command cmdChat;
	/**
	 * Command for changing current screen to camera.
	 */
	private Command cmdCamera;
	/**
	 * Command for changing current screen to joystick.
	 */
	private Command cmdJoystick;
	/**
	 * Command for sending message in the chat form.
	 */
	private Command cmdSendMessage;
	/**
	 * Command for writing new message in the chat form.
	 */
	private Command cmdNewMessage;
	/**
	 * List object containing list of devices.
	 */
	private List devicesList;
    /**
     * TextBox object containing message text for MobileImagineConnectorServer.
     */
	private TextBox txtMessageBox;


	/**
	 * Constructs new MobileImagineConnectorClient class.
	 * Creates all commands and forms used in this class.
	 */
	public MobileImagineConnectorClient() {
		this.formMain = new Form("Mobile Imagine Connector");
		formMain.append("Start by searching for devices.\n\n");

		formChat = new Form("Chat");

        cmdBack = new Command("Back", Command.BACK, 0);
        formChat.addCommand(cmdBack);

        cmdNewMessage = new Command("New message", Command.SCREEN, 0);
        formChat.addCommand(cmdNewMessage);
        formChat.setCommandListener(this);

		// these three commands will be added to the main form when
		// the connection to the server is established
		cmdChat = new Command("Chat", Command.SCREEN, 0);
		cmdCamera = new Command("Camera", Command.SCREEN, 0);
		cmdJoystick = new Command("Joystick", Command.SCREEN, 0);

        cmdOBEXclient = new Command("Devices", Command.SCREEN, 0);
        formMain.addCommand(cmdOBEXclient);

		cmdExit = new Command("Exit", Command.EXIT, 0);
		formMain.addCommand(cmdExit);
        formMain.setCommandListener(this);
	}

	/**
	 * From MIDlet:
	 * Called when MIDlet is started.
     * Gets the Display instance, sets current screen to the main form and
     * initializes instance of ObexClient.
	 */
	protected void startApp() throws MIDletStateChangeException {
		display = Display.getDisplay(this);
		setMIDletDisplay(null, formMain);

		obexClient = new ObexClient(this);
	}

    /**
     * From MIDlet:
     * Called to signal the MIDlet to terminate, UNUSED in this application.
     * @param unconditional 	If true, then the MIDlet has to be unconditionally 
     * terminated and all resources has to be released.
     */
	protected void destroyApp(boolean unconditional) throws MIDletStateChangeException {
	}

	/**
	 * From MIDlet:
	 * Called when MIDlet is paused, UNUSED in this application.
	 */
	protected void pauseApp() {
	}

    /**
     * Exits MIDlet.
     */
    private void exitMIDlet() {
    	setMIDletDisplay(null, null);
        try {
			destroyApp(true);
		} catch (MIDletStateChangeException e) {
			e.printStackTrace();
		}
        notifyDestroyed();
    }
 
    /**
     * Switches a current displayable in a display. The Display instance is 
     * taken from getDisplay() method. This method is used by all actions 
     * in the design for switching displayable.
     * @param alert 			The Alert which is temporarily set to the display.
     * @param nextDisplayable 	The Displayable to be set.
     */
    void setMIDletDisplay(Alert alert, Displayable nextDisplayable) {
    	if (alert == null) {
            display.setCurrent(nextDisplayable);
        } else {
            display.setCurrent(alert, nextDisplayable);
        }
    }

    /**
     * Shows alert with specified title and text.
     * @param title 			Title of alert.
     * @param message 			Text of alert.
     * @param type 				Type of Alert
     * @param timeout			Number of milliseconds the alert is shown or until user clicks button if Alert.FOREVER is used.
     * @param nextDisplayable 	Next displayable.
     */
    private void showAlert(String title, String message, AlertType type, int timeout, Displayable nextDisplayable) {
        alert = new Alert(title, message, null, type);
        alert.setTimeout(timeout);

       	setMIDletDisplay(alert, nextDisplayable);
    }        

    /**
     * Adds a text to the main form.
     * @param message	The text to be printed into the main form.
     */
    void printToForm(String message){
    	formMain.append(message + "\n\n");
    }

    /**
     * Returns an instance of Display. Used by Camera class.
     * @return	The instance of Display.
     */
    Display getDisplay(){
    	return display;
    }

    /**
     * Returns an instance of Camera. Used by ObexClient. 
     * @return	The instance of Camera.
     */
    Camera getCamera(){
		if (camera == null)
			camera = new Camera(this);
    	return camera;
    }

    /**
     * Returns main form.
     * @return	Main form of this MIDlet.
     */
    Form getFormMain(){
    	return formMain;
    }

    /**
     * Returns chat form chat.
     * @return	Chat form.
     */
    Form getFormChat(){
    	return formChat;
    }

    /**
     * Switches the current screen to Camera's view finder.
     */
    void switchToCamera(){
		if (camera == null)
			camera = new Camera(this);
		camera.showCamera();
    }

    /**
     * Switches the current screen to Camera's form with captured pictures.
     */
    void switchToImages(){
		if (camera == null)
			camera = new Camera(this);
	    setMIDletDisplay(null, camera.getFormImages());
    }

    /**
     * Switches the current screen to Joystick.
     */
    void switchToJoystick(){
    	if (joystick == null)
    		joystick = new Joystick(this);
	    setMIDletDisplay(null, joystick);
    }

    /**
     * Returns the command for searching for devices.
     * @return	The command for searching for devices.
     */
    Command getCommandDevices(){
    	return cmdOBEXclient;
    }

    /**
     * Returns an instance of ObexClient.
     * @return	The instance of ObexClient.
     */
    ObexClient getObexClient(){
    	return obexClient;
    }

    /**
     * Returns an initialized instance of List component.
     * @return 	The initialized component instance.
     */
    List getDevicesList() {
        if (devicesList == null) {
            devicesList = new List("Found devices", List.IMPLICIT);

            devicesList.addCommand(cmdBack);

            devicesList.setCommandListener(this);
        }
        return devicesList;
    }

    /**
     * Returns an initialized instance of TextBox component.
     * @return	The initialized component instance.
     */
    private TextBox getTxtMessageBox() {
        if (txtMessageBox == null) {
            txtMessageBox = new TextBox("Message", null, 1000, TextField.ANY);
            cmdSendMessage = new Command("Send", Command.ITEM, 0);
 
            txtMessageBox.addCommand(cmdSendMessage);
            txtMessageBox.addCommand(cmdBack);
            txtMessageBox.setCommandListener(this);
        }
        return txtMessageBox;
    }    

    /**
     * Sends the message with message type to the MobileImagineConnectorServer.
     * The message type can be either message from chat or pressed key from an instance
     * of Joystick.
     * @param messageString		The message to be sent to MobileImagineConnectorServer.
     * @param messageType		The type of message, either 1 for message from chat, or 2 for message from an instance of Joystick.
     */
    void sendMessageToServer(String messageString, int messageType){
    	obexClient.PutMessage(messageString.getBytes(), messageType);
    }

    /**
     * Sends the array of bytes with message type to the MobileImagineConnectorServer.
     * The message type can be only 3 for captured image from an instance of Camera.
     * @param messageBytes		The array of bytes to be sent to MobileImagineConnectorServer.
     * @param messageType		The type of message, can be only 3 for captured image from an instance of Camera.
     */
    void sendMessageToServer(byte[] messageBytes, int messageType){
    	obexClient.PutMessage(messageBytes, messageType);
    }

    /**
     * From CommandListener:
     * Called by a system to indicate that a command has been invoked on a 
     * particular displayable.
     * @param command 		The Command that was invoked.
     * @param displayable 	The Displayable where the command was invoked.
     */
    public void commandAction(Command command, Displayable displayable) {
    	// displayable is the main form
		if (displayable == formMain){
			if (command == cmdExit){
				exitMIDlet();
			} else if (command == cmdOBEXclient){
				// removing command "Devices"
				formMain.removeCommand(cmdOBEXclient);
				// clearing old devices data
		        getDevicesList().deleteAll();
				obexClient.getDevices();
				showAlert("Info", "Make sure the server is running before selecting a device!", AlertType.INFO, 5000, formMain);
			} else if (command == cmdChat){
				setMIDletDisplay(null, formChat);
			} else if (command == cmdCamera){
				switchToCamera();
			} else if (command == cmdJoystick){
				switchToJoystick();
			}
		// displayable is the list of found devices
		} else if (displayable == devicesList){
			if (command == List.SELECT_COMMAND) {
				getDevicesList().append("selecting devide", null);
				//starting search for Messenger service on selected device
				RemoteDevice currentDevice = 
					(RemoteDevice) obexClient.getDevicesVector().elementAt(
							getDevicesList().getSelectedIndex());
				try {
					printToForm("Choosen device: " + currentDevice.getFriendlyName(false));
				} catch (IOException e) {
				}
				obexClient.startDiscoverService(currentDevice);
				setMIDletDisplay(null, getFormMain());
			} else if (command == cmdBack){
				setMIDletDisplay(null, getFormMain());
			}
		// displayable is the chat form
		} else if (displayable == formChat){
			if (command == cmdBack){
				setMIDletDisplay(null, getFormMain());
			} else if (command == cmdNewMessage) {
                setMIDletDisplay(null, getTxtMessageBox());
			}
		// displayable is the text box (for typing of new message)
		} else if (displayable == txtMessageBox){
			if (command == cmdBack){
				setMIDletDisplay(null, formChat);
			} else if (command == cmdSendMessage){
				sendMessageToServer(txtMessageBox.getString(), ObexClient.MESSAGE_TYPE_TEXT);
                txtMessageBox.setString("");
                setMIDletDisplay(null, getFormChat());
			}
		}
	}

    /**
     * Called when the connection with MobileImagineConnectorServer is started or interrupted.
     * If the connection is started, the three commands for showing chat form, camera or joystick
     * form are added to the main form, if the connection is interrupted, these commands are removed.
     * @param connected		If true, the three commands for showing chat form, camera or joystick form are added to the main form, if false, these commands are removed.
     */
    void setConnected(boolean connected){
		if (connected){
			formMain.addCommand(cmdChat);
			formMain.addCommand(cmdCamera);
			formMain.addCommand(cmdJoystick);
        } else {
			formMain.removeCommand(cmdChat);
			formMain.removeCommand(cmdCamera);
			formMain.removeCommand(cmdJoystick);
        }
    }
}