package org.wakeland.pedro;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
 
import java.util.Enumeration;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Vector;
 
import javax.microedition.io.Connector;
import javax.microedition.lcdui.Alert;
import javax.microedition.lcdui.Displayable;
import javax.microedition.lcdui.Form;
 
import javax.bluetooth.BluetoothStateException;
import javax.bluetooth.DataElement;
import javax.bluetooth.DeviceClass;
import javax.bluetooth.DiscoveryAgent;
import javax.bluetooth.DiscoveryListener;
import javax.bluetooth.LocalDevice;
import javax.bluetooth.RemoteDevice;
import javax.bluetooth.ServiceRecord;
import javax.bluetooth.UUID;
 
import javax.obex.ClientSession;
import javax.obex.HeaderSet;
import javax.obex.Operation;
import javax.obex.ResponseCodes;
 
/**
 * 
 * J2ME class for starting a Bluetooth connection over OBEX protocol with 
 * MobileImagineConnectorServer (running on PC). 
 * 
 * <p>Original sample code of OBEX client downloaded from
 * <a href="http://wiki.forum.nokia.com/index.php/Using_Bluetooth_OBEX_Client_in_Java_ME">Nokia wiki forum</a>.</p>
 * 
 * @author Peter (Pedro) Nosal, pedron4@gmail.com
 * @version 1.0
 *
 */
public class ObexClient implements DiscoveryListener {

	/**
	 * Instance of main class - MIDlet.
	 */
	private MobileImagineConnectorClient midlet;
    /**
    * Flag indicating that server service is found.
    */
    private boolean isServiceFound;
    /**
     * Timer object for periodic sending GET requests to MobileImagineConnectorServer.
     */    
    private Timer timer;
    /**
    * The DiscoveryAgent for the local Bluetooth device.
    */
    private DiscoveryAgent discoveryAgent;
    /**
    * Keeps track of the devices found.
    */
    private Vector devicesVector;
    /**
    * Keeps track of the transaction ID returned from searchServices.
    */
    private int transactionID;
    /**
    * Describes ClassID of Messenger Service (UUID).
    */
    private final String serviceClassIDstring;
    /**
     * Provides methods for OBEX requests.
     */
    private ClientSession connectionSession;
    /**
     * Constant for message type - text.
     */
    public final static int MESSAGE_TYPE_TEXT = 1;
    /**
     * Constant for message type - pressed button.
     */
    public final static int MESSAGE_TYPE_BUTTON = 2;
    /**
     * Constant for message type - image.
     */
    public final static int MESSAGE_TYPE_IMAGE = 3;


    /**
     * Constructs new ObexClient class.
     * @param midlet	Main class - MIDlet.
     */
    public ObexClient(MobileImagineConnectorClient midlet) {
    	this.midlet = midlet;
        try {
            discoveryAgent = LocalDevice.getLocalDevice().getDiscoveryAgent();
        } catch (BluetoothStateException ex) {
            // Failed to get LocalDevice DiscoveryAgent Object.
            // It means that the Bluetooth system could not be initialized
        	this.midlet.printToForm("Application was unable to locate Bluetooth on you phone.");
        }
 
        isServiceFound = false;
        serviceClassIDstring = "49C41BB556414702B914EFEE734AF38B";
        devicesVector = new Vector();
        connectionSession = null;
    }
 
    /**
     * Switches a current displayable in a display. The Display instance is 
     * taken from getDisplay() method. This method is used by all actions 
     * in the design for switching displayable.
     * @param alert 			The Alert which is temporarily set to the display.
     * @param nextDisplayable 	The Displayable to be set.
     */
    public void switchDisplayable(Alert alert, Displayable nextDisplayable) {
    	midlet.setMIDletDisplay(alert, nextDisplayable);
    }

    /**
     * Returns a Vector of found devices.
     * @return	The Vector of found devices.
     */
    public Vector getDevicesVector(){
    	return devicesVector;
    }

    /**
     * Starts device discovery.
     */
    public void getDevices(){
        devicesVector.removeAllElements();
        try {
        	// starting discovery devices process
        	midlet.printToForm("Starting devices search...");
        	discoveryAgent.startInquiry(DiscoveryAgent.GIAC, this);
        	isServiceFound = false;
        } catch (BluetoothStateException e) {
        	midlet.printToForm("Start device discovery error!");
        }
    }
 
    /**
    * From DiscoveryListener:
    * Called when a device was found during an inquiry. An inquiry
    * searches for devices that are discoverable. The same device may
    * be returned multiple times. If the found device's CoD indicates
    * the device is computer, the device is added to the Vector of found
    * devices.
    * @see DiscoveryAgent#startInquiry
    * @param btDevice 	The device that was found during the inquiry.
    * @param cod 		The service classes, major device class, and minor device class of the remote device being returned.
    */
    public void deviceDiscovered(RemoteDevice btDevice, DeviceClass cod) {
    	// checking if device is computer
    	if (cod.getMajorDeviceClass() == 256){
            // adding new device object to Vector object
	        devicesVector.addElement(btDevice);
    	}
    }
 
    /**
    * From DiscoveryListener:
    * Called when service(s) are found during a service search.
    * This method provides the array of services that have been found.
    * Trying to connect if the right service (according to the UUID) has been found.
    * @param transID 			The transaction ID of the service search that is posting the result.
    * @param serviceRecords 	A list of services found during the search request.
    */    
    public void servicesDiscovered(int transID, ServiceRecord[] serviceRecords){
        for (int i = 0; i < serviceRecords.length; i++) {
            DataElement rcvObject;
 
            // getting ServiceClassID (0x0001) as DataElement object
            rcvObject = serviceRecords[i].getAttributeValue(0x0001); 
            if (rcvObject != null) {
                // rcvObject presents as enumeration of DataElement
                // getting enumeration...
                Enumeration enumeration = (Enumeration)rcvObject.getValue();
 
                for (Enumeration e = enumeration; e.hasMoreElements() ;) {
                    DataElement subDataElement = (DataElement)e.nextElement();
                    // ...and converting every element of enumeration to UUID
                    UUID uuid = (UUID) subDataElement.getValue();
 
                    // comparing needed UUID with present
                    if (serviceClassIDstring.equals(uuid.toString())) {
                        printToForm("Service found!");
                        isServiceFound = true;
                        // if UUIDs are the same then trying to connect
                        // getting the connection string for connection
                        String connectionString = serviceRecords[i].getConnectionURL(ServiceRecord.NOAUTHENTICATE_NOENCRYPT, false);
 
                        try {
                            // getting the client connection object
                            connectionSession = (ClientSession) Connector.open(connectionString, Connector.READ_WRITE);
                        } catch (IOException ex) {
                        	printToForm("IOException during getting connection object.");
                        }
 
                        // trying to connect
                        printToForm("Connecting to server...");
                        HeaderSet headerSet = null;
                        try {
                        	// sending CONNECT request
                        	headerSet = connectionSession.connect(null);
                        } catch (IOException ex) {
                        	printToForm("IOException during connecting to the MobileImagineConnectorServer.");
                        }

                        try {
                        	if (headerSet.getResponseCode() == ResponseCodes.OBEX_HTTP_OK) {
                        		printToForm("Connection successful!");
                        	} else {
                        		printToForm("Error during requesting to connectto the MobileImagineConnectorServer. code: " + headerSet.getResponseCode());
                        		connectionSession.close();
                        		return;
                        	}
                        } catch (IOException ex) {
                        	printToForm("IOException during requesting to connect to the MobileImagineConnectorServer.");
                        }
 
                        // connection established
                        midlet.setConnected(true);
                        // TimerTask object creating for periodic 
                        // asynchronous sending GET requests to server
                        ObexClientTimerTask timerTask = new ObexClientTimerTask(this);
                        // timer starting
                        timer = new Timer();
                        timer.schedule((TimerTask)timerTask, 1000);
                        return;
                    }
                }
            }
        }
    }
 
    /**
    * From DiscoveryListener:
    * The following method is called when a service search is completed or
    * was terminated because of an error.
    * @param transID 		The transaction ID identifying the request which initiated the service search.
    * @param respCode 		The response code which indicates the status of the transaction; guaranteed to be one of the aforementioned only.
    */
    public void serviceSearchCompleted(int transID, int respCode) {
        if (transactionID == transID) {
            switch(respCode) {
                case DiscoveryListener.SERVICE_SEARCH_COMPLETED:
                    if (isServiceFound) {
                        printToForm("Service discovering complete!");
                    } else {
                        printToForm("Service discovering incomplete! Service not found!");
                    }
                    break;
                case DiscoveryListener.SERVICE_SEARCH_TERMINATED:
                    printToForm("Services discovering terminated!");
                    break;
                case DiscoveryListener.SERVICE_SEARCH_ERROR:
                    printToForm("Sevice discovering error!");
                    break;
                default:
                    break;
            }
            transactionID = -1;
            return;
        } else {
            // if transactionID are not the same as transID
            // it's an error
            printToForm("Error during connecting to the MobileImagineConnectorServer.");
        }
        printToForm("Discover Service Complete...");
    }
 
    /**
    * From DiscoveryListener:
    * Called when a device discovery transaction is completed. 
    * @param discType	The type of request that was completed.
    */
    public void inquiryCompleted(int discType){
        switch(discType) {
            case INQUIRY_COMPLETED:
                int devcesCount = devicesVector.size();
                if (devcesCount == 0) {
                    printToForm("There are no devices! Try to do the search again later...");
                    midlet.getFormMain().addCommand(midlet.getCommandDevices());
                    return;
                }
 
                // before creating new elements in devicesList we must clear old values
                midlet.getDevicesList().deleteAll();
 
                for (int i = 0; i < devcesCount; i++) {
                    String deviceName = "Noname";
                    try {
                        deviceName = ((RemoteDevice) devicesVector.elementAt(i)).getFriendlyName(false);
                        if (deviceName.equals(""))
                        	deviceName = "Noname";
                    } catch (IOException e) {
                        printToForm("IOException on getting friendly name.");
                        return;
                    }
                    midlet.getDevicesList().append(deviceName, null);
                }
                printToForm("Discovering devices complete!");
 
                // switching to devicesList where we can choose the device
                // for service discovery
                switchDisplayable(null, midlet.getDevicesList());
                midlet.getFormMain().addCommand(midlet.getCommandDevices());
                break;
            case INQUIRY_TERMINATED:
                printToForm("Discover devices terminated!");
                break;
            case INQUIRY_ERROR:
                printToForm("Discover devices error!");
                break;
        }
    }
 
    /**
     * Adds a text to the MIDlet's main form.
     * @param strPrint 	String to add to the MIDlet's main form.
     * @see Form#append(java.lang.String) 
     * @see Form
     */
    public void printToForm(String strPrint) {
    	midlet.printToForm(strPrint);
    }

    /**
     * Starts service discovery on current device, used in commandAction().
     * @param currentDevice 	The RemoteDevice. Current device for searching.
     */
    public void startDiscoverService(RemoteDevice currentDevice) {
        if (currentDevice == null) {
            throw new NullPointerException("Can't get RemoteDevice object!");
        }
        // we can use UUID for OBEX protocol, which is 0x0008
        // but we can also use UUID of our server application,
        // which is saved in the variable serviceClassIDstring
        UUID[] searchList = new UUID[1];
        searchList[0] = new UUID(serviceClassIDstring, false);

        // start searching services on current device
        // and get transaction ID of the service search
        try {
            transactionID = discoveryAgent.searchServices(null, searchList, currentDevice, this);
            printToForm("Starting service search...");
        } catch (BluetoothStateException e) {
            //Failed to start the search on this device, try another
            //device.
        	printToForm("exception on searching services");
        	printToForm(e.toString());
        }
    }
    /**
     * Used by ObexClientTimerTask for sending GET request to server
     * checking response and restarting the timer.
     */
    public void GetMessageOnTimer() {
        GetMessage();
        ObexClientTimerTask timerTask = new ObexClientTimerTask(this);
        timer.schedule((TimerTask)timerTask, 1500);
    }
    /**
     * Used for creating PUT request to OBEX Server.
     * Sending either text message or image in jpeg format as array of bytes to server.
     *
     * WARNING!!! 
     * If the length of the message to be sent is less than 1024 bytes,
     * the server calls the onDelete() method, not the onPut() method. To make sure
     * the onPut() method is being called, message with length less than 1024 bytes
     * is filled by 0's (zeros) to have 1024 bytes.
     * 
     * @param messageBytes	Array of bytes to sent to MobileImagineConnectorServer.
     * @param messageType	The type of message.
     */
    public void PutMessage(byte[] messageBytes, int messageType) {
    	// we can use only one Operation object at one time
        // it mean's that when we send PUT request we can't
        // send GET request
        synchronized(connectionSession) {

        	// creating header for PUT request
            HeaderSet head = connectionSession.createHeaderSet();
 
            try {
                // setting up type and length of message inside header
                head.setHeader(HeaderSet.TYPE, "BtObexMessage");
                head.setHeader(HeaderSet.LENGTH, new Long(1024));
            } catch (IllegalArgumentException ex) {
            	printToForm("IllegalArgumentException during sending data to MobileImagineConnectorServer.");
                return;
            }
 
            Operation operation;
            try {
                // creating new Operation object for current connection
                operation = connectionSession.put(head);
            } catch (IOException ex) {
            	printToForm("IOExeption during sending data to MobileImagineConnectorServer.");
                return;
            }
 
            // issue the request
            // open the output stream to put the object to it
            OutputStream out = null;
            try {
            	out = operation.openOutputStream();
            } catch (IOException e) {
            	printToForm("IOException during sending data to MobileImagineConnectorServer.");
            }

            // sending either text (chat) or button pressed in joystick
            if (messageType != MESSAGE_TYPE_IMAGE){
            	// the message will be sent in the following format:
            	// <messageType><messageLength>*<actualMessage>
            	// if length of this message is less than 1024, we add 0's (zeros) at the end
            	// so that message will be 1024 bytes long

            	// print the message to MIDlet's main form
            	String messageString = "";
            	for (int i = 0; i < messageBytes.length; i++)
            		messageString += (char)messageBytes[i];
            	if (messageType == MESSAGE_TYPE_TEXT)
            		midlet.getFormChat().append("sent: " + messageString + "\n");

            	String messageStart = messageType + String.valueOf(messageBytes.length) + "*";

            	// calculating how many zeros we have to add
            	int addZeros = 1024 - messageBytes.length - messageStart.length();
            	String messageEnd = "0";
            	for (int i = 0; i < addZeros; i++)
            		messageEnd += "0";

            	// send the object to the server
            	try {
                	// sending message type, message length and asterisk
					out.write(messageStart.getBytes());
	            	// sending actual message
					out.write(messageBytes);
	            	// if we need to send zeros
	            	if (addZeros > 0){
						out.write(messageEnd.getBytes());
	            	}
				} catch (IOException e) {
				}
	        // sending image
            } else {
            	midlet.printToForm("Sending image to server.");

            	String messageStart = messageType + "";
            	try {
            		out.write(messageStart.getBytes());
					out.write(messageBytes);
				} catch (IOException e) {
	            	printToForm("IOException while sending image to the MobileImagineConnectorServer.");
				}
            }
        	// end the transaction
			try {
				out.flush();
			} catch (IOException e) {
				printToForm("IOException during sending data to the MobileImagineConnectorServer.");
			}
            // close the stream
            try {
            	out.close();
            } catch (IOException e) {
				printToForm("IOException during sending data to the MobileImagineConnectorServer.");
            }
            // close the operation
            try {
            	operation.close();
            } catch (IOException e) {
				printToForm("IOException during sending data to the MobileImagineConnectorServer.");
            }
        }
    }        

    /**
     * Used for creating GET request to OBEX Server.
     * Getting text message from server.
     */
    public void GetMessage(){
    	// we can use only one Operation object at one time
    	// it mean's that when we send GET request we can't
    	// sending PUT request
    	synchronized(connectionSession){
    		// header creating for GET request
	    	HeaderSet reqHeader = connectionSession.createHeaderSet();
	    	// setting the type of the object to retrieve
	    	reqHeader.setHeader(HeaderSet.TYPE, "BtObexMessage");
	    	reqHeader.setHeader(HeaderSet.LENGTH, new Long(0));
	
	    	// getting headers
	    	Operation operation = null;
			try {
				operation = connectionSession.get(reqHeader);
			} catch (IOException e) {
				printToForm("Error: connection with server lost!");
				midlet.setConnected(false);
			}
	
			// opening input stream
	    	InputStream inStream = null;
			try {
				inStream = operation.openInputStream();
			} catch (IOException e) {
				printToForm("IOException during received data from the MobileImagineConnectorServer.");
			}
	
	    	if (inStream != null) {
	            // reading data
	            StringBuffer buf = new StringBuffer();
	            int data;
	            try {
					while ((data = inStream.read()) != -1)
					    buf.append((char)data);
				} catch (IOException e1) {
					printToForm("IOException during received data from the MobileImagineConnectorServer.");
				}

				// getting response code
	    		try {
					operation.getResponseCode();
				} catch (IOException e) {
					printToForm("IOException during received data from the MobileImagineConnectorServer.");
				}

				// closing operation
	    		try {
					operation.close();
				} catch (IOException e) {
					printToForm("IOException during received data from the MobileImagineConnectorServer.");
				}

				// parsing message
				if (buf.length() != 0){
					parseMessage(buf.toString());
				}
	    	}
    	}
    }

    /**
     * Parses the message that came from server.
     * Looks up for the first letter and then decides what to do.
     * Set of commands (depending on what the first letter is):
     * <ul>
     * <li>a</li>: receive regular message to chat
     * <li>b</li>: switch screen to main
     * <li>c</li>: switch screen to chat
     * <li>d</li>: switch screen to camera
     * <li>e</li>: switch screen to images
     * <li>f</li>: switch screen to joystick
     * <li>g</li>: switch screen to camera and capture picture
     * <li>h</li>: send picture back to server (if there is one)
     * <li>j</li>: switch screen to camera, capture picture and send it back to server
     * </ul>
     * @param message	The message that came from server.
     */
    private void parseMessage(String message){
		switch (message.charAt(0)){
		case 'a':
			midlet.getFormChat().append("received: " + message.substring(1) + "\n");			
			break;
		case 'b':
			switchDisplayable(null, midlet.getFormMain());
			break;
		case 'c':
			switchDisplayable(null, midlet.getFormChat());
			break;
		case 'd':
			midlet.switchToCamera();
			break;
		case 'e':
			midlet.switchToImages();
			break;
		case 'f':
			midlet.switchToJoystick();
			break;
		case 'g':
			midlet.switchToCamera();
			midlet.getCamera().capture();
			break;
		case 'h':
			if (midlet.getCamera().isCaptured())
				midlet.getCamera().sendImageToServer();
			break;
		case 'j':
			midlet.switchToCamera();
			midlet.getCamera().capture();
			while (!midlet.getCamera().isCaptured()){
			}
			midlet.getCamera().sendImageToServer();
			break;
		}
    }
}